/*

 DESCRIPTION:
 Include file for GPIBCON.

 Copyright 2021 Brian Cornell.

 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in
 the Software without restriction, including without limitation the rights to
 use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 of the Software, and to permit persons to whom the Software is furnished to do
 so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

 DEPENDENCIES:  none

 VERSION:  5.1

 MODIFICATION HISTORY:

 01 Mar 21, BWC, minor revision 5.1
 10 Oct 19, BWC, fifth release.
 18 Sep 19, BWC, minor revision 4.1.
 20 Apr 19, BWC, fourth release.
 29 Mar 19, BWC, third release.
 24 Mar 19, BWC, second release.
 13 Feb 19, BWC, first release.


 */


#define VERMAJOR        5
#define VERMINOR        1
#define VERBUILD        5

#define TRUE            1
#define FALSE           0

#define SYS_EPOCH_YEAR  1900            //Epoch for C library
#define LOG_EPOCH_YEAR  2000            //Epoch for logger HW

#define FMODE_GPIB      "r+"            //Need r/w
#define FMODE_CONF      "r"             //Read only
#define FMODE_LOG_FIL   "w"             //Open & truncate, write only
#define FMODE_CFG_FIL   "a"             //OPen for append, create if not exist

#define BASE_TEN        10

#define POS_ARG_PRG     0
#define DSP_POS_VAL_R   3
#define DSP_POS_VAL_C   1
#define DSP_POS_CMD_R   15              //Command prompt
#define DSP_POS_CMD_C   2               //Command prompt
#define DSP_POS_UPD_R   DSP_POS_CMD_R   //Display uppdate progress indicator
#define DSP_POS_UPD_C   (DSP_POS_CMD_C - 1)
#define DSP_POS_RSP_R   16              //Command response
#define DSP_POS_RSP_C   0               //Command response
#define DSP_POS_EXT_R   24              //Quit requested from console
#define DSP_POS_EXT_C   79              //Quit requested from console
#define DSP_POS_PRG_R   24              //Program information
#define DSP_POS_PRG_C   0               //Program information
#define DSP_POS_MKL_R   18              //Mark status label, see comments for constant FMT_MKL_MARK
#define DSP_POS_MKL_C   1               //Mark status label
#define DSP_POS_MKS_R   DSP_POS_MKL_R   //Mark status value
#define DSP_POS_MKS_C   7               //Mark status value

#define DSP_POS_LOG_R   19              //Data logging active info
#define DSP_POS_LOG_C   0               //Data logging active info

//Execution delay constants
#define DLY_S_INT_C     0               //# S between initialization commands
#define DLY_NS_INT_C    20000000        //# nS between initialization commands
#define DLY_S_RD_C      0               //# S between read commands
#define DLY_NS_RD_C     1000000         //# nS between read commands
#define DLY_S_DUP_C     0               //# S between display updates
#define DLY_NS_DUP_C    900000000       //# nS between display updates
#define DLY_S_OFT_C     1               //# S between off commands
#define DLY_NS_OFT_C    0               //# nS between off commands
#define DLY_S_OFF_C     0               //# S between off commands
#define DLY_NS_OFF_C    20000000        //# nS between off commands
#define DLY_MS_MAX_INIT 999             //Maximum init mS delay
#define DLY_INT_NS_CONV 1000000         //Init delay multiplier for nanosleep (1mS)

//Ansi terminal control & characters
#define ANSIESC         "\e["
#define ANSICLRSCR      "2J"
#define ANSICPOS        'H'
#define ANSIELINE       'K'
#define ANSISAVE        "\e7"
#define ANSIRESTORE     "\e8"
#define ACP_SZ_STR      4
#define ACP_MAX_ROW     25
#define ACP_MAX_COL     80
#define CHR_CR          0x0D
#define CHR_LF          0x0A
#define CHR_NULL        0x00
#define CHR_COLON       ':'
#define CHR_SEMICOLON   ';'
#define CHR_SPACE       ' '
#define CHR_C_LC        'c'
#define CHR_C_UC        'C'

//Dimensions
#define SIZ_BUF_CFG     32
#define SIZ_BUF_CMD     32
#define SIZ_BUF_HLP     64
#define SIZ_BUF_RSP     64              //GPIB response buffer
#define SIZ_BUF_VAR     64              //Should be >= largest buffer in gcmd
#define SIZ_BUF_FILE    256
#define SIZ_ARGC_MIN    2               //Minimum # arguments
#define SIZ_BUF_CON     70              //User console (command)
#define SIZ_CMD_FLAG    2               //Size of command line flag
#define SIZ_ARY_FLD     2               //Array holding float values of math fields
#define SIZ_ARY_NULM    2               //Size of NULL string message array

//Array addresses
#define ARY_FLD_FLD1    0               //First math field
#define ARY_FLD_FLD2    1               //Second math field

//Program strings
#define STR_PRGNAME     "gpibcon"       //Program name
#define STR_EXT_CFG     ".conf"         //Config extension
#define STR_ERROR       "ERR"           //Error
#define STR_UPDATE      "|/-|\\-"        //Display update monitor
#define STR_ON          "ON "
#define STR_OFF         "OFF"

//Print formats
#define FMT_INT_WR      "%s\n"          //End initialization command with NL
#define FMT_INT_DELAY   "\nDev %d delay %dmS"  //Initialization delay message
#define FMT_INT_MATCH   "\nSending %s to dev %d, adr %d, to match response to %s"
#define FMT_INT_CMD     "\nSending command %s to dev %d, adr %d"
#define FMT_ADR_CHG     "%s %d\n"       //Change GPIB address
#define FMT_OFF_WR      "%s\n"          //End off command with NL
#define FMT_DSP_WR      "%s\n"          //End display command with NL
#define FMT_DSP_VAL     "%s%s\t"        //Display value format
#define FMT_DMP_HDR     "\n---%s section---\n"
#define FMT_DMP_INT     "dev=%d, adr=%d, cmd=%s, cresp=%s, delay=%d\n"
#define FMT_DMP_RD      "dev=%d, adr=%d, cmd=%s, to=%d\n"
#define FMT_DMP_ADR     "dev=%d, adr=%d, cmd=%s\n"
#define FMT_DMP_CFG     "dev=%d, adr=%d, cfg=%s, cmd=%s, usrv=%d, crsp=%d, hlp=%s\n"
#define FMT_DMP_DSP     "%d: dev=%d, adr=%d, label=%s, cmd=%s, row=%d, rltx=%d, info=%s\n"
#define FMT_DMP_MTH     "\top=%c, f1 | constant=%d | %f, f2 | constant=%d | \
%f, precision=%d\n"
#define FMT_DMP_OFF     "dev=%d, adr=%d, cmd=%s\n"
#define FMT_DMP_END     "\n# records=%d, max=%d\n"
#define FMT_CMD_PR      ": "
#define FMT_RSP_PR      "=>"
#define FMT_RSP_NOV     "Value required"
#define FMT_EXC_CMD     "%s%s\n"
#define FMT_RSP_IOE     "device io error"
#define FMT_EXC_RSP     "%s\n\n\n%s"
#define FMT_RSP_INV     "Invalid command"
#define FMT_HLP_DCM     "\n++++++++++Dictonary Commands++++++++++\n"
#define FMT_HLP_ICM     "\n++++++++++Internal Commands++++++++++\n"
#define FMT_HLP_TXT     "%s: %s\n"
#define FMT_HLP_END     "\n\nPress ENTER to resume."
#define FMT_ERR_STDERR  "\n%s:  %s, errno=%d."
#define FMT_TRM_SIGNAL  "\n%s terminating on signal %d.\n"
#define FMT_PRG_DESC    "\nGPIB console interface, v%d.%d build %d.  \
(c) 2021 Brian Cornell.\n"
#define FMT_PRG_SYNTX   "Syntax:  %s (flags) <device> (<device>).  Optional flags:"
#define FMT_PRG_CMDL    "\n\t%s - %s"
#define FMT_PRG_EXIT    "\n\n"
#define FMT_MTH_RESULT  "%06.*f"
#define FMT_LOG_MSG     "Writing data to %s"
#define FMT_LOG_OUT     "\n%u,%2.2d/%2.2d/%d,%2.2d:%2.2d:%2.2d,%d,\"%s\""
#define FMT_LOG_FLD     ",%s"
#define FMT_LOG_HDR     "Record,Date,Time,Mark,Comment"
#define FMT_CFG_OUT     "\n%u,%2.2d/%2.2d/%d,%2.2d:%2.2d:%2.2d,%s,%s,%s"
#define FMT_CFG_HDR     "Record,Date,Time,Command,Value,Response"
#define FMT_STR_UPDATE  "%c"
#define FMT_MKS_MARK    "%s"
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Constant FMT_MKL_MARK contains two leading new lines as a HACK to force
 * display to correct row when printed from function scrSetup().  Function
 * ansiCursorPos does not set the row correctly:  consistently two rows short
 * (e.g. passing row argument of 18 prints positions cursor to row 16).
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define FMT_MKL_MARK    "\n\nMark:"

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Constants for gcmd objects.  Device number - POS_XXXX_DEV is always the
 * first field and can be addressed with the non-record specific constant
 * POS_ALL_DEV.
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Type gcmd object types
#define REC_TYP_READ    0               //Command to read value
#define REC_TYP_INIT    1               //Initialization command
#define REC_TYP_CFG     2               //User entered configuration commands
#define REC_TYP_DSP     3               //Display commands
#define REC_TYP_OFF     4               //Shutdown commands
#define REC_TYP_ADR     5               //Change address command

//Field positions common to all object types
#define POS_ALL_DEV     1               //Device number
#define POS_ALL_ADR     2               //Address number

//Field positions for read records
#define POS_REA_DEV     1               //GPIB device number
#define POS_REA_ADR     2               //Instrument address
#define POS_REA_CMDC    3               //GPIB command
#define POS_REA_TOUT    4               //Device read timeout, mS

//Field positions for off records
#define POS_OFF_DEV     1               //GPIB device number
#define POS_OFF_ADR     2               //Instrument address
#define POS_OFF_CMDC    3               //GPIB command

//Field positions for change address records
#define POS_ADR_DEV     1               //GPIB device number
#define POS_ADR_ADR     2               //Instrument address
#define POS_ADR_CMDC    3               //GPIB command

//Field positions for init records
#define POS_INT_DEV     1               //GPIB device number
#define POS_INT_ADR     2               //Instrument address
#define POS_INT_CMDC    3               //GPIB command
#define POS_INT_CRSP    4               //Compare command response
#define POS_INT_DELAY   5               //Delay initialization

//Field positions for cfg records
#define POS_CFG_DEV     1               //GPIB device number
#define POS_CFG_ADR     2               //Instrument address
#define POS_CFG_USRC    3               //User (console) command name
#define POS_CFG_CMDC    4               //GPIB (machine) command name
#define POS_CFG_PARM    5               //Command parameter required
#define POS_CFG_CRSP    6               //Read command response
#define POS_CFG_HELP    7               //Help string

//Field positions for dsp records
#define POS_DSP_DEV     1               //GPIB device number
#define POS_DSP_ADR     2               //Instrument address
#define POS_DSP_LABL    3               //Display label
#define POS_DSP_CMDC    4               //GPIB (machine) command name
#define POS_DSP_ROW     5               //Display row
#define POS_DSP_RLTX    6               //Remove leading text from response
#define POS_DSP_MATH    7               //Math operator
#define POS_DSP_FLD1    8               //1st operand field
#define POS_DSP_FLD2    9               //2nd operand field
#define POS_DSP_PREC    10              //Decimal precision for calculated fields
#define POS_DSP_HELP    11               //Help string

//Config file command prefixes
#define REC_NAM_INIT    "init:"         //Initialization
#define REC_NAM_READ    "read:"         //Read
#define REC_NAM_CFG     "cfg:"          //Configuration
#define REC_NAM_DSP     "dsp:"          //Display
#define REC_NAM_OFF     "off:"          //Shutdown
#define REC_NAM_ADR     "addr:"         //Address

//Record command tags
#define DSP_TAG_CALC    "$calc$"        //If found in command field indicates calculated field
#define INT_TAG_DELAY   "$delay$"       //If found in command field indicates delay

//Math operators
#define MATH_OP_BAD     0               //Invalid
#define MATH_OP_ADD     1               //Add
#define MATH_OP_SUB     2               //Subtract
#define MATH_OP_MUL     3               //Multiply
#define MATH_OP_DIV     4               //Divide

//Math operator symbols array
#define MATH_ARY_OPS    " +-*/"

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * These constants dimension an array of gcmd pointers that point to the first &
 * last objects of each dictionary type (e.g. init, display, etc.).
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define DICT_INIT_F     0               //First init record
#define DICT_INIT_L     1               //Last init record
#define DICT_CONF_F     2               //First config record
#define DICT_CONF_L     3               //Last config record
#define DICT_DISP_F     4               //First display record
#define DICT_DISP_L     5               //Last display record
#define DICT_OFF_F      6               //First off record
#define DICT_OFF_L      7               //Last off record
#define SIZ_ARY_IDX     8               //Size of array

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * These constants dimension gcmd object array used for read commands and GPIB
 * device stream pointers.  Since only one read command is allowed per device
 * there is a 1-1 relationship and a common set of constants is used.
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define SIZ_ARY_DEV     2               //Device steam & read commands
#define ARY_FLD_DEV1    0               //First GPIB device / read command
#define ARY_FLD_DEV2    1               //Second GPIB device / read command

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * These constants dimension gcmd object array used for addr commands.  Since
 * only one address command is allowed per device there is are two address
 * records.
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define SIZ_ARY_ADR     2               //Array of address commands
#define ARY_FLD_ADR1    0               //First GPIB device addr command
#define ARY_FLD_ADR2    1               //Second GPIB device addr command
#define VAL_ADR_INIT    999             //Starting invalid address for setDevAddr

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Dictionary object definition & array dimension
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define SIZ_ARY_DCT     500             //Static size of dictionary

typedef struct {
    unsigned int devn;                  //GPIB device number from config file
    unsigned int addr;                  //GPIB instrument address
    FILE *devf;                         //GPIB device stream
    int   type;                         //Object type
    int   usrv;                         //type=cfg - value req, type=dsp - row, type=init or read - delay n mS
    int   crsp;                         //type=cfg get command response, type=dsp remove leading text
    int   math;                         //type=dsp - see MATH_OP_XXX constants
    int   prec;                         //Decimal precision for result of calculated fields
    int   fld[SIZ_ARY_FLD];             //Display fields to perform calc on, offset-1 or 0 to use fval
    float fval[SIZ_ARY_FLD];            //Constant to use for calculated field instead of display field
    void *prev;                         //Pointer to previous object of type
    void *next;                         //Pointer to next object of type
    char  cfg[SIZ_BUF_CFG];             //type=cfg - user command, type=dsp - label, type=init - compare rsp
    char  cmd[SIZ_BUF_CMD];             //Machine version of command
    char  hlp[SIZ_BUF_HLP];             //type=cfg Help, type=dsp label info
    char  rsp[SIZ_BUF_RSP];             //GPIB response
} gcmd;


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Constants for cmd objects
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Command object
typedef struct {
    const char *str;
    const char *desc;
    char       *value;
} cmd;

//Internal command strings
#define CMD_A_CLS       0
#define CMD_A_HELP      1
#define CMD_A_MARK      2
#define CMD_A_QUIT      3

const cmd Icmd[] = {
    {"cls","Redraw screen"},
    {"help","Display command help"},
    {"mark","Toggle mark log file on/off"},
    {"quit","Quit program"}
};

#define SIZ_A_ICMD      sizeof(Icmd) / (sizeof(const char *) * 3)


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Command line processing
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
//Flag bit mapping
#define FLG_NONE            0x0000
#define FLG_BIT_SET         0x0001
#define FLG_DIFF_CONF       0x0001
#define FLG_DUMP_CONF       0x0002
#define FLG_FORCE_DSP       0x0004
#define FLG_CFG_WRITE       0x0008
#define FLG_LOG_FILE        0x0010
#define FLG_RESET_DSP       0x0020
#define FLG_DSP_UPDATE      0x0040

//Flag array position
#define FLG_A_DIFF_CONF     0
#define FLG_A_DUMP_CONF     1
#define FLG_A_FORCE_DSP     2
#define FLG_A_CFG_WRITE     3
#define FLG_A_LOG_FILE      4
#define FLG_A_RESET_DSP     5
#define FLG_A_DSP_UPDATE    6

cmd Flags[] = {
    {"-c","use different config file"},
    {"-d","dump config file & exit"},
    {"-f","force display reset always"},
    {"-g","log config (console) commands to file"},
    {"-l","log display data to <file>"},
    {"-r","reset display on ERR"},
    {"-u","update display every <mS>"}
};

#define SIZ_A_FLG       sizeof(Flags) / (sizeof(const char *) * 3)


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Return codes & messages
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define RC_SUCCESS      0
#define RC_NOARG        1
#define RC_BADARG       2
#define RC_BADCFG       3
#define RC_DEVOFAIL     4
#define RC_SIGERROR     5
#define RC_INITFAIL     6
#define RC_THREAD       7
#define RC_SIGEXIT      8
#define RC_LOGFAIL      9
#define RC_CMDFAIL      10

const char *retCode[] = {
    "success",
    "no arguments",
    "bad or missing arguments",
    "problem with config file",
    "can't open device",
    "error installing signal handler",
    "error initializing device",
    "problem starting display thread",
    "CTRL-C or TERM signal received",
    "data log file error",
    "Cfg (command) log file error"
};


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Function declarations
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int main(int, char *[]);
int cmdProc(void);
int loadCfg(char *, int, gcmd *[], gcmd *, gcmd *, gcmd *);
void dumpCfg(gcmd *[], gcmd *, gcmd *, gcmd *);
int gpibInit(gcmd *, gcmd *);
int gpibOff(gcmd *);
void gpibDsp(void *);
int strnl(unsigned char *, int);
int strlnshift(unsigned char *, int);
int strlrtrim(unsigned char *, int);
void sigHandler(int);
const char *getCfgVar(char *, int);
void ansiClrScr(void);
void ansiEraseLine(void);
void ansiCursorPos(unsigned char, unsigned char);
void ansiCursorSav(void);
void ansiCursorRst(void);
void scrSetup(void);
void gpibCalc(gcmd *);
int setDevAddr(gcmd *);
int gpibRead(gcmd *, int);
int gpibLogc(char *, char *, char *, char *);



//End file.
